/*
  The purpose of this function is to get an image from an internet location, resize it and save it to S3.
*/

/*
IAM POLICY NOTES FOR S3 (Avoiding "AccessDenied: Access Denied")
The Lambda function usually uses a role that you point to in order for the Lambda function to execute.  If you do not already have a policy in
place to handle interacting with objects in specific locations on S3, you should do so.  Below is an example policy you can create in the IAM.
Once created you can attach the policy to the role you have attached to the Lamda function.  In this example, the ability to get and put objects
is allowed for the bucket "yourbucket" and for a SPECIFIC folder "images".  This means that the role by using the function will NOT be able to 
get or put images in any other bucket or folder, other than "yourbucket/images/":
{
	"Version": "2012-10-17",
	"Statement": [
		{
		"Effect": "Allow",
		"Action": [
			"s3:GetObject",
			"s3:PutObject"
		],
		"Resource": [
			"arn:aws:s3:::yourbucket/images/*"
		]
		}
	]
}
*/

/*
CUSTOM FUNCTIONS TO SIMPLIFY HANDLING RETRIEVAL DATA
*/
function retrievalSetExensionName(data) {
	/* Set Staging Name for Image and Extension */
	var result = true;
	if (data.indexOf(".") > -1) {
		tmp_extension = [];
		tmp_extension = data.split(".");
		var imgExtension = tmp_extension[(tmp_extension.length - 1)].toLowerCase();
		if (imgExtension === "jpeg") { imgExtension === "jpg"; }
		if (imgExtension === "jpg" || imgExtension === "jpeg" || imgExtension === "png") {
			/* Append extension to staging name */
			imgStagingName = imgStagingName + "." + imgExtension;
			retFileExtension = imgExtension;
		}
		else { result = false; }
	}
	else { result = false; }
	return result;
}

function retrievalParseHeaders(data) {
	/* Parse Response Header Data */
	var result = true;
	var haveValidContent = 0;
	try {
		retServer = data["server"].toString();
		retDate = data["date"].toString();
		retContentType = data["content-type"].toString();
		retContentLength = Number(data["content-length"]);
		retConnection = data["connection"].toString();
		retLastModified = data["last-modified"].toString();
		retAcceptRanges = data["accept-ranges"].toString();
		retVary = data["vary"].toString();
	}
	catch(e) { result = false; }
	/* Validate content type for jpg or png files.  This is not the only way to validate source data, but a good start. */
	if (retContentType.toLowerCase().indexOf("image/jpeg") > -1) { haveValidContent = 1; }
	if (retContentType.toLowerCase().indexOf("image/jpg") > -1) { haveValidContent = 1; }
	if (retContentType.toLowerCase().indexOf("image/png") > -1) { haveValidContent = 1; }
	if (haveValidContent === 0) { retFlag = 1; }
	return result;
}

function retrievalStageFile(callback) {
	/* Retrieves and stages a file to S3 (in this case, an image) */
	results = {}; retBody = [];
	/* Fork based on connection type */
	var tmp_protocol = sourceURL.split(":")[0].toLowerCase();
	if (tmp_protocol === "https") {
		https.get(sourceURL, function(result) {
			/* Error */
			result.on('error', function(e) {
				results = {
					"status": 0,
					"message": e.message
				};
				callback(results);
				return;
			});

			/* Collect Response Stream Data */
			result.on('data', function(chunk) { retBody.push(chunk); });

			/* Req / Res End. */
			result.on('end', function() {
				/* Parse Response Header Data */
				var tmpStatus = retrievalParseHeaders(result.headers);
				if (tmpStatus) {
					if (retFlag === 0) {
						/* Response data found and of correct file type */
						/* Save the file to the S3 bucket */
						retBucketParams = {
							Bucket: storageS3Bucket,
							Key: imgStagingName,
							Body: Buffer.concat(retBody)
						};
						s3.putObject(retBucketParams, function(err, data) {
							if (err) {
								results = {
									"status": 0,
									"message": "Error with S3: " + err.stack
								};
								callback(results);
								return;
							}
							else {
								results = {
									"status": 1,
									"message": "Saved to S3."
								};
								callback(results);
								return;
							}
						});
					}
					else {
						/* Response data but wrong type */
						results = {
							"status": 0,
							"message": "Response data was found but of the wrong type - " + retContentType
						};
						callback(results);
						return;
					}
				}
				else {
					/* No response data */
					results = {
						"status": 0,
						"message": "No response data was found.  The URL may be incorrect or your request to the image is being blocked."
					};
					callback(results);
					return;
				}
			});
		});
	}
	else if (tmp_protocol === "http") {
		http.get(sourceURL, function(result) {
			/* Error */
			result.on('error', function(e) {
				results = {
					"status": 0,
					"message": e.message
				};
				callback(results);
				return;
			});

			/* Collect Response Stream Data */
			result.on('data', function(chunk) { retBody.push(chunk); });

			/* Req / Res End. */
			result.on('end', function() {
				/* Parse Response Header Data */
				var tmpStatus = retrievalParseHeaders(result.headers);
				if (tmpStatus) {
					if (retFlag === 0) {
						/* Response data found and of correct file type */
						/* Save the file to the S3 bucket */
						retBucketParams = {
							Bucket: storageS3Bucket,
							Key: imgStagingName,
							Body: Buffer.concat(retBody)
						};
						s3.putObject(retBucketParams, function(err, data) {
							if (err) {
								results = {
									"status": 0,
									"message": "Error with S3: " + err.stack
								};
								callback(results);
								return;
							}
							else {
								results = {
									"status": 1,
									"message": "Saved to S3."
								};
								callback(results);
								return;
							}
						});
					}
					else {
						/* Response data but wrong type */
						results = {
							"status": 0,
							"message": "Response data was found but of the wrong type - " + retContentType
						};
						callback(results);
						return;
					}
				}
				else {
					/* No response data */
					results = {
						"status": 0,
						"message": "No response data was found.  The URL may be incorrect or your request to the image is being blocked."
					};
					callback(results);
					return;
				}
			});
		});
	}
	else {
		/* Connection protocol not supported */
		results = {
			"status": 0,
			"message": "Connection protocol not supported."
		};
		callback(results);
		return;
	}
}

function retrievalResizeImg(callback) {
	/* Resize the staged image on S3 */
	/* The following is a modification of the example provided by AWS at docs.aws.amazon.com/lambda/latest/dg/with-s3-example-deployment-pkg.html */
	results = {};
	async.waterfall([
		function download(next) {
			// Download the image from S3 into a buffer.
			s3.getObject({
				Bucket: storageS3Bucket,
				Key: imgStagingName
			},
			next);
		},
		function transform(response, next) {
			gm(response.Body).size(function(err, size) {
				// Infer the scaling factor to avoid stretching the image unnaturally.
				var scalingFactor = Math.min(
					imgResizeX / size.width,
					imgResizeY / size.height
				);
				var width  = scalingFactor * size.width;
				var height = scalingFactor * size.height;
				// Transform the image buffer in memory.
				this.resize(width, height).toBuffer(retFileExtension, function(err, buffer) {
					if (err) {
						next(err);
					}
					else {
						next(null, response.ContentType, buffer);
					}
				});
			});
		},
		function upload(contentType, data, next) {
			// Stream the transformed image to a different S3 bucket.
			s3.putObject({
				Bucket: storageS3BucketFinal,
				Key: imgResizeSaveName,
				Body: data,
				ContentType: contentType
			},
			next);
		},
	], function (err) {
		if (err) {
			results = {
				"status": 0,
				"message": err
			};
		}
		else {
			results = {
				"status": 1,
				"message": "Image resized."
			};
		}
		callback(results);
	}
	);
}

/*
AWS COMPONENTS
*/
var AWS = require("aws-sdk");
var http = require("http");
var https = require("https");
var s3 = new AWS.S3();

/*
LOCAL COMPONENTS FOUND IN "node_modules" folder
*/
var async = require("async");				/* https://www.npmjs.com/package/async */
var gm = require("gm").subClass({ imageMagick: true });	/* https://www.npmjs.com/package/gm */
var util = require("util");				/* https://www.npmjs.com/package/aws-node-util */

/*
FILE RETRIEVAL HANDLING VARIABLES
*/
var retBody = [], retBucketParams = {}, retFileExtension = "";
var retFlag = 0;
var retServer = "";
var retDate = "";
var retContentType = "";
var retContentLength = 0;
var retConnection = "";
var retLastModified = "";
var retAcceptRanges = "";
var retVary = "";

/*
S3 - RESIZING VARIABLES
*/
var storageS3Bucket = "stagingfolder/images";			/* Staging area for image retrieved via http/s */
var storageS3BucketFinal = "finalfolder/images";		/* Location of the resized, final images */
var imgStagingName = "rawimage";				/* Staging name of the source image without extension (extension added separately) */
var imgLargeName = "bigimage";					/* Name of the resized, final large image */
var imgSmallName = "smallimage";				/* Name of the resized, final small image */
var imgLargeX = 200;						/* Pixel width of the resized, final large image */
var imgLargeY = 200;						/* Pixel height of the resized, final large image */
var imgSmallX = 75;						/* Pixel width of the resized, final small image */
var imgSmallY = 75;						/* Pixel height of the resized, final small image */
var sourceURL = "http://www.yoursite.com/yourimage.jpg";	/* URL location of the source image to resize */
var imgResizeX = 0;
var imgResizeY = 0;
var imgResizeSaveName = "";

/*
MAIN FUNCTION
*/
exports.handler = function(event, context) {
	/* Validate extension of the source image and set the staging name */
	if (retrievalSetExensionName(sourceURL)) {
		/* Get file and save it to S3 bucket */
		var stage = retrievalStageFile(function(stageResponse) {
			if (Number(stageResponse.status) === 1) {
				/* Resize the staged image to create what will be used on the website */
				imgLargeName = imgLargeName + "." + retFileExtension;
				imgSmallName = imgSmallName + "." + retFileExtension;
				/* STEP 1: Resize and save large image of original */
				imgResizeX = imgLargeX;
				imgResizeY = imgLargeY;
				imgResizeSaveName = imgLargeName;
				var resizeda = retrievalResizeImg(function(resizedaResponse) {
					if (Number(resizedaResponse.status) === 1) {
						/* Resizing for large image completed */
						/* STEP 2: Resize and save small image of original */
						imgResizeX = imgSmallX;
						imgResizeY = imgSmallY;
						imgResizeSaveName = imgSmallName;
						var resizedb = retrievalResizeImg(function(resizedbResponse) {
							if (Number(resizedbResponse.status) === 1) {
								context.done(
									null,
									{
									"Result": "Make pancakes and waffles."
									}
								);
							}
							else {
								/* Resizing for small image could not be completed */
								context.done(
									null,
									{
									"Result": "Error: Resize operation for small image failed.  " + resizedbResponse.message
									}
								);
							}
						});
					}
					else {
						/* Resizing for large image could not be completed */
						context.done(
							null,
							{
							"Result": "Error: Resize operation for large image failed.  " + resizedaResponse.message
							}
						);
					}
				});
			}
			else {
				/* There was a problem with retrieving and storing the image to the staging location */
				context.done(
					null,
					{
					"Result": "Error: " + stageResponse.message
					}
				);
			}
		});
	}
	else {
		/* Filetype not supported */
		context.done(
			null,
			{
			"Result": "Image filetype must be jpg, jpeg or png format."
			}
		);
	}
};